/*-----------------------------------------------------------*/
/* DÉBUT DU CODE DE BASE. NE PAS MODIFIER.                   */
/*-----------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>


/* Pointeur global réservé pour le tri. Ne pas utiliser. */
int* ordre_global_reserve = NULL;

/* Fonction de comparaison pour le tri. Ne pas utiliser. */
int comp_ordre_glob(int* u, int* v) {
  int c;
  if (ordre_global_reserve[*u] < ordre_global_reserve[*v]) {
    c = -1;
  } else if (ordre_global_reserve[*u] > ordre_global_reserve[*v]) {
    c = 1;
  } else if (*u < *v) {
    c = -1;
  } else if (*v < *u) {
    c = 1;
  } else {
    c = 0;
  }
  return c;
}


/* Les fonctions qui suivent peuvent être utilisées dans vos programmes
   si c'est utile. */

/* Prend en entrée un tableau a de n entiers,
   et un tableau ordre.
   Si ordre contient m éléments, les éléments de a doivent être
   entre 0 et m-1.
   Trie a dans l'ordre croissant pour l'ordre "<<" défini par
   "p << q lorsque ordre[p] < ordre[q], ou (ordre[p] = ordre[q] et p < q)"
   où "<" est l'ordre habituel sur les entiers.
   La fonction modifie directement le tableau a,
   plutôt que de renvoyer le tableau trié. */
/* comlexité en temps : O(n*log(n)) /*
void trie_tableau(int* a, int n, int* ordre) {
  ordre_global_reserve = ordre;
  qsort(a, n, sizeof(int), (int (*) (const void *, const void*)) &comp_ordre_glob);
  ordre_global_reserve = NULL;
  return;
}


/* Affiche une matrice a de taille n * n */
void affiche_matrice(int** a, int n) {
  printf("/");          
  for (int j = 0; j < n; j++) {
    printf("  ");
  }
  printf(" \\\n");
  for (int i = 0; i < n; i++) {
    printf("|");
    for (int j = 0; j < n; j++) {
      printf(" %d", a[i][j]);
    }
    printf(" |\n");
  }
  printf("\\");
  for (int j = 0; j < n; j++) {
    printf("  ");
  }
  printf(" /\n");
  return;
}


/* On stockera dans Un la table des u_n précalculés */
int* Un;

/* On précalculera u_n jusqu'à n = max_n - 1 */
const int max_n = 3000000;


/*-----------------------------------------------------------*/
/* FIN DU CODE DE BASE.                                      */
/*-----------------------------------------------------------*/


/* utilitaires */
/* pour afficher les erreurs */
void erreur(char* e) {
  printf("Erreur: %s\n", e);
  exit(1);
}

/* alloue la mémoire pour une matrice n*n */
int** alloue_matrice(int n) {
  int** a = malloc(n * sizeof(int*));
  if (!a) { erreur("alloue_matrice malloc a"); }
  for (int i = 0; i < n; i++) {
    a[i] = malloc(n * sizeof(int));
    if (! a[i]) { erreur("alloue_matrice malloc a[i]"); }
  }
  return a;
}

/* free la matrice a de taille n*n */
void free_matrice(int n, int** a) {
  for (int i = 0; i < n; i++) {
    free(a[i]);
  }
  free(a);
  return;
}



/*----------------*/
/*  Question 7    */
/*----------------*/

/* remplit la table Un */
void precalcul_un(int u0) {
  Un[0] = u0;
  for (int i = 1; i < max_n; i++) {
    Un[i] = (569 * Un[i-1]) % 2424259;
    assert(569 * Un[i-1] > -1);
  }
  return;
}


/* renvoie u_n */
int un(int n) {
  assert(n > -1 && n < max_n);
  return Un[n];
}


/* calcule A_n,m,p */
int** anmp(int n, int m, int p) {
  int** a = alloue_matrice(n);
  
  /* remplit d'abord les i <= j */
  for (int i = 0; i < n; i++) {
    for (int j = i; j < n; j++) {
      if ((un(n*n + 89*m + 71*p + n*i + j) % p) < m) {
        a[i][j] = 1;
      } else {
        a[i][j] = 0;
      }
    }
  }

  /* puis les i > j */
  for (int i = 0; i < n; i++) {
    for (int j = 0; j < i; j++) {
      a[i][j] = a[j][i];
    }
  }

  return a;
}
   

/* calcul de la somme pour la q7 */
int q7(int n, int m, int p) {
  int** a = anmp(n,m,p);
  int s = 0;
  for (int i=0; i < n; i++) {
    for (int j=0; j < n; j++) {
      s = s + ((n*i + j) * a[i][j]);
    }
  }
  free_matrice(n, a);
  return s % 10000;
}


/* affichage du résultat */
void print_q7(char* num, int n, int m, int p) {
  printf("7.%s %d\n", num, q7(n, m, p));
  return;
}

void qq7() {
  print_q7("a", 3, 1, 2);
  print_q7("b", 20, 3, 4);
  print_q7("c", 100, 3, 4);
  print_q7("d", 1000, 1, 1000);
  printf("\n");
  return;
}
 
 
/*----------------*/
/*  Question 8    */
/*----------------*/
 
/* calcule la largeur de bande de a de taille n * n */
int largeurbande(int n, int** a) {
  int l = 0;
  for (int i = 0; i < n; i++) {
    for (int j = 0; j < n; j++) {
      if (a[i][j] != 0 && abs(i-j) > l) {
        l = abs(i-j);
      }
    }
  }
  return l;
}


int q8(int n, int m, int p) {
  int** a = anmp(n,m,p);
  int l = largeurbande(n, a);
  free_matrice(n, a);
  return l;
}

/* affichage du résultat */
void print_q8(char* num, int n, int m, int p) {
  printf("8.%s %d\n", num, q8(n, m, p));
  return;
}

void qq8() {
  print_q8("a", 10, 1, 10);
  print_q8("b", 100, 1, 100);
  print_q8("c", 200, 1, 1000);
  printf("\n");
  return;
}




/*----------------*/
/*  Question 9    */
/*----------------*/

/* calcule la table des degrés de tous les sommets de a */
int* degres(int n, int** a) {
  int* deg = malloc(n * sizeof(int));
  if (!deg) { erreur("degres malloc deg"); }
  for (int i = 0; i < n; i++) {
    deg[i] = 0;
    for (int j = 0; j < n; j++) {
      if (a[i][j] != 0) {
        deg[i] = deg[i] + 1;
      }
    }
  }
  return deg;
}



/* parcours en largeur de a, renvoie l'ordre de parcours des sommets */
int* parcours(int n, int** a) {

  /* stockage des sommets dans un tableau */
  int* sommets = malloc(n * sizeof(int));
  if (!sommets) { erreur("parcours malloc sommets"); }
  for (int i = 0; i < n; i++) {
    sommets[i] = i;
  }

  /* calcul des degrés */
  int* deg = degres(n, a);

  /* tri par degré croissant */
  trie_tableau(sommets, n, deg); 
  
  /* file implémentée par un tableau et deux indices */
  int* f = malloc(n * sizeof(int)); // on n'aura que n entiers dans la file
  int f_in = 0; // indice où écrire pour enfiler
  int f_out = 0; // indice où écrire pour défiler. f_in == f_out ssi f est vide

  /* tableau pour stocker l'ordre de visite -- stocke vus en même temps */
  int* ordre = malloc(n * sizeof(int));
  if (!ordre) { erreur("parcours malloc ordre"); }
  for (int i = 0; i < n; i++) {
    ordre[i] = -1; // -1 si i pas encore vu
  }

  int o = 0; // ordre du sommet suivant

  for (int si = 0; si < n; si++) {
    int s = sommets[si]; // on parcourt les sommets par degré croissant
    if (ordre[s] == -1) { // on ne fait rien si s déjà vu
      /* à ce stade s est le plus petit sommet non déjà vu */
      ordre[s] = o; // s vu
      o++;
      f[f_in] = s; // enfile s
      f_in++;

      while (f_out < f_in) {
        int s0 = f[f_out]; // defile s0
        f_out++;

        /* calcul des voisins non vus (par degré croissant) */
        for (int j = 0; j < n; j++) {
          int v = sommets[j];
          if ((a[s0][v] != 0) && (ordre[v] == -1)) {
            ordre[v] = o; // v vu
            o++;
            f[f_in] = v; // enfile v
            f_in++;
          }
        }
      }
    }
  }
  
  assert(o == n); // on doit avoir vu tout le monde
  assert(f_in == f_out); // f doit être vide
  
  free(f);
  free(deg);
  free(sommets);
  return ordre;
}


/* renumérote les sommets de a (n * n) suivant ordre (ordre = sigma^-1)*/
int** renumerote(int n, int** a, int* ordre) {
  int** ra = alloue_matrice(n);
  for (int i = 0; i < n; i++) {
    for (int j = 0; j < n; j++) {
      ra[ordre[i]][ordre[j]] = a[i][j];
    }
  }
  return ra;
}


int q9(int n, int m, int p) {
  int** a = anmp(n, m, p);
  int* ordre = parcours(n, a);
  int** ra = renumerote(n, a, ordre);
  free(ordre);
  free_matrice(n, a);
  int s = 0;
  for (int i = 0; i < n; i++) {
    for (int j = 0; j < n; j++) {
      s = (s + ((n*i + j) * ra[i][j])) % 10000;
    }
  }
  free_matrice(n, ra);
  return (s % 10000);
}


/* affichage du résultat */
void print_q9(char* num, int n, int m, int p) {
  printf("9.%s %d\n", num, q9(n, m, p));
  return;
}

void qq9() {
  print_q9("a", 10, 1, 10);
  print_q9("b", 100, 1, 100);
  print_q9("c", 1000, 1, 1000);
  print_q9("d", 1000, 1, 10);
  printf("\n");
  return;
}




/*-----------------------*/
/* Tous les résultats    */
/*-----------------------*/

int main(int argc, char** argv) {

  Un = malloc(max_n * sizeof(int));
  if (!Un) { erreur("allocation Un"); }

  /* précalcul de tous les u_n */
  precalcul_un(1571);
  
  qq7();
  qq8();
  qq9();
  
  return 0;
}
