(* -------------- *)
(* Question 1     *)
(* -------------- *)

(* u0~ utilisé dans le sujet *)
let u0 : int = 1571

(* on précalcule toutes les valeurs jusqu'à max_n dans une table *)
let max_n : int = 2000000
let un_tab : int array = Array.make (max_n + 1) 0
let f u = (569 * u) mod 2424259

let _ =
  un_tab.(0) <- u0;
  for i = 1 to max_n do
    un_tab.(i) <- f un_tab.(i-1);
  done

let un (n:int) : int =
  assert (n <= max_n);
  un_tab.(n)

let q1 (n:int) : int =
  (un n) mod 10000


(* affichage et valeurs numériques *)
let print_q1 (num:string) (n:int) =
  Printf.printf "1.%s %d\n" num (q1 n);
  flush stdout

let qq1 () =
  print_q1 "a" 1;
  print_q1 "b" 128;
  print_q1 "c" 2024;
  print_q1 "d" 123456;
  Printf.printf "\n";
  flush stdout



(* -------------- *)
(* Question 2     *)
(* -------------- *)



(* représentation de la grille par une matrice de booléens *)
type case = int * int
type grille = bool array array

(* quelques fonctions pour crééer une grille vide,
   obtenir la taille d'une grille, lire ou inverser l'état d'une case,
   lister les cases allumées *)
let vide (n:int) : grille = Array.make_matrix n n false
let taille (g:grille) : int = Array.length g
let est_allume (g:grille) ((i, j):case) : bool = g.(i).(j)
let inverse (g:grille) ((i, j):case) : unit = g.(i).(j) <- not (g.(i).(j))
let cases_allumees (g:grille) : case list =
  let l = ref [] in
  Array.iteri
    (fun i -> Array.iteri (fun j b -> if b then l := (i,j) :: !l)) g;
  !l

(* calcul de la somme pour les valeurs numériques demandées *)
let somme_ensemble (c:case list) : int =
  let s = List.fold_left (fun s (i,j) -> s + i + j) 0 c in
  s mod 10000

(* appuie sur la case p de g *)
let appuie (g:grille) (p:case) : unit =
  let n = taille g in
  let (i,j) = p in
  let c = [i, j; i, j+1;  i-1, j; i, j-1; i+1, j] in
  let c = List.filter (fun (x,y) -> x < n && y < n && x >= 0 && y >= 0) c in
  List.iter (inverse g) c

(* calcul de G_n,m,p *)
let gnmp (n:int) (m:int) (p:int) : grille =
  let g = vide n in
  for i=0 to n-1 do
    for j=0 to n-1 do
      let pp = (un (n*n + 89*m + 71*p + n*i + j)) mod p in
      if pp < m then 
        inverse g (i,j);
    done;
  done;
  g

let q2 (n:int) (m:int) (p:int) : int =
  let g = gnmp n m p in
  somme_ensemble (cases_allumees g)


(* affichage et valeurs numériques *)
let print_q2 (num:string) (n:int) (m:int) (p:int) : unit =
  Printf.printf "2.%s %d\n" num (q2 n m p);
  flush stdout

let qq2 () =
  print_q2 "a" 3 1 4;
  print_q2 "b" 10 3 4;
  print_q2 "c" 100 2 5;
  Printf.printf "\n";
  flush stdout




(* -------------- *)
(* Question 3     *)
(* -------------- *)

let distance (g:grille) (h:grille) : int =
  let n = taille g in
  let d = ref 0 in
  for i=0 to n-1 do
    for j=0 to n-1 do
      if est_allume g (i,j) <> est_allume h (i,j) then
        incr d;
    done;
  done;
  !d

let q3 (n:int) (m:int) (p:int) (m':int) (p':int) : int =
  let g = gnmp n m p in
  let h = gnmp n m' p' in
  distance g h


(* affichage et valeurs numériques *)
let print_q3 (num:string) (n:int) (m:int) (p:int) (m':int) (p':int) : unit =
  Printf.printf "3.%s %d\n" num (q3 n m p m' p');
  flush stdout

let qq3 () =
  print_q3 "a" 3 1 4 1 2;
  print_q3 "b" 10 3 4 1 2;
  print_q3 "c" 100 2 5 3 7;
  Printf.printf "\n";
  flush stdout



(* -------------- *)
(* Question 4     *)
(* -------------- *)

(*  variation de distance g h lorsque l'on appuie sur la case p de h *)
let delta (g:grille) (h:grille) (p:case) : int =
  let n = taille g in
  assert (n = taille h);
  let (i,j) = p in
  let c = [i, j; i, j+1;  i-1, j; i, j-1; i+1, j] in
  let c = List.filter (fun (x,y) -> x < n && y < n && x >= 0 && y >= 0) c in
  List.fold_left
    (fun x p ->
      x + (if (est_allume g p) = (est_allume h p) then 1 else -1))
    0 c


let forcebrute (n:int) (g:grille) : case list =
  let h = vide n in (* grille de travail *)

  (* calcule le meilleur choix possible pour les cases <= p,
     sachant que c est le choix actuel pour les cases > p,
     qu'il atteint une distance d et une somme s,
     et que le meilleur choix jusque là est cmin qui atteint dmin, smax *)
  let rec aux p c d s cmin dmin smax =
    let (i,j) = p in
    if i = -1 then (* plus de case à traiter :
                      on regarde si on a trouvé un meilleur ensemble *)
      (if d < dmin || (d = dmin && s > smax) then (c, d, s)
       else (cmin, dmin, smax))
    else
      let pred = (* prédécesseur de p pour l'ordre lexicographique *)
        if j = 0 then (i-1, n-1) else (i, j-1) in
      (* on essaie sans appuyer sur p *)
      let c1, d1, s1 = aux pred c d s cmin dmin smax in

      (* si on appuie sur p : la distance et la somme actuelles changent *)
      let dd = delta g h p in
      let ds = i + j in
      appuie h p;
      let c2, d2, s2 = aux pred (p::c) (d+dd) ((s+ds) mod 10000) c1 d1 s1 in
      appuie h p;
      (c2, d2, s2)
  in

  (* on appelle aux en partant d'une grille vide et de la plus grande case *)
  let d = distance g h in
  let c, _, _ = aux (n-1,n-1) [] d 0 [] d 0 in
  c

let q4 (n:int) (m:int) (p:int) : int =
  let g = gnmp n m p in
  let c = forcebrute n g in
  somme_ensemble c


(* affichage et valeurs numériques *)
let print_q4 (num:string) (n:int) (m:int) (p:int) : unit =
  Printf.printf "4.%s %d\n" num (q4 n m p);
  flush stdout

let qq4 () =
  print_q4 "a" 3 1 2;
  print_q4 "b" 4 1 4;
  print_q4 "c" 4 3 4;
  Printf.printf "\n";
  flush stdout


(* -------------- *)
(* Question 5     *)
(* -------------- *)

let glouton (g:grille) : case list =
  let n = taille g in
  let h = vide n in (* grille de travail *)
  let d, c = ref (distance h g), ref [] in (* distance et c courants *)
  let dmin, cmin = ref !d, ref !c in (* meilleurs distance et c vus *)

  (* table pour stocker les cases choisies. *)
  let deja_pris = Array.make_matrix n n false in

  for k = 1 to n * n do (* on va ajouter toutes les cases -> n^2 tours *)
    (* on parcourt toutes les cases pour trouver la meilleure *)
    let ddmin = ref 6 in (* le plus bas delta trouvé jusque là *)
    let pmin = ref (n,n) in (* la case qui le réalise *)
    for i=0 to n-1 do
      for j=0 to n-1 do
        let dd = delta g h (i, j) in
        if dd <= !ddmin && not deja_pris.(i).(j) then (* trouvé mieux *)
          (ddmin := dd; pmin := (i, j))
      done;
    done;

    (* on choisit la meilleure case trouvée *)
    d := !d + !ddmin;
    c := !pmin :: !c;
    deja_pris.(fst !pmin).(snd !pmin) <- true;
    appuie h !pmin;

    (* on met à jour la meilleure distance vue, si nécessaire *)
    if !d < !dmin then (dmin := !d; cmin := !c);
  done;
  !cmin

let q5 (n:int) (m:int) (p:int) : int =
  let g = gnmp n m p in
  let c = glouton g in
  somme_ensemble c


(* affichage et valeurs numériques *)
let print_q5 (num:string) (n:int) (m:int) (p:int) : unit =
  Printf.printf "5.%s %d\n" num (q5 n m p);
  flush stdout

let qq5 () =
  print_q5 "a" 3 3 4;
  print_q5 "b" 20 2 5;
  print_q5 "c" 40 5 7;
  Printf.printf "\n";
  flush stdout



(* -------------- *)
(* Question 6     *)
(* -------------- *)

(* Tout d'abord trois fonctions pour traduire le problème vers le système
   linéaire *)

(* calcul du vecteur V_G *)
let vg (n:int) (g:grille) : bool array =
  let v = Array.make (n*n) false in
  List.iter (fun (i,j) -> v.(i*n + j) <- true) (cases_allumees g);
  v

(* calcul de la matrice M *)
let matrice_lumiere (n:int) : bool array array =
  let nn = n * n in
  let m = Array.make nn (Array.make 0 false) in
  for i = 0 to nn - 1 do 
    m.(i) <- Array.make nn false;
  done;
  for i=0 to (n-1) do 
    for j=0 to (n-1) do 
      let l = [(i, j); (i, j+1); (i-1, j); (i, j-1); (i+1, j)] in
      let l = List.filter 
                (fun (x,y) -> x >= 0 && y >= 0 && x < n && y < n) l
      in
      List.iter (fun (x,y) -> m.(i * n + j).(x * n + y) <- true) l;
    done;
  done;
  m

(* calcul de C à partir de V_C *)
let cases_vc (n:int) (vc:bool array) : case list =
  let l = ref [] in
  Array.iteri (fun p vcp -> if vcp then l := (p / n, p mod n):: !l) vc;
  !l


(* Puis le pivot de Gauss *)

(* addition dans Z/2Z *)
let xor (a:bool) (b:bool) : bool = a <> b

(* ajoute la ligne i2 à la ligne i1 à la fois dans a et b *)
let ajoute a b i1 i2 = 
  Array.iteri (fun j x -> a.(i1).(j) <- xor x a.(i1).(j)) a.(i2);
  b.(i1) <- xor b.(i2) b.(i1)

(* renvoie une copie d'une matrice carrée a *)
let copie a =
  let n = Array.length a in
  let aa = Array.make_matrix n 0 false in
  Array.iteri (fun i ai -> (aa.(i) <- Array.copy ai)) a;
  aa

(* échange les lignes i1 et i2 dans a *)
let echange a i1 i2 =
  let c = a.(i1) in
  a.(i1)<- a.(i2);
  a.(i2)<- c

(* pivot de Gauss *)
(* a matrice n * n, b vecteur n, résout a x = b *)
(* dans le cas où a est de rang n *)
let pivot (n:int) (a:bool array array) (b:bool array) =
  (* copies de a, b pour ne pas toucher aux originaux *)
  let a = copie a in
  let b = Array.copy b in

  (* a est supposée de rang n : on sait qu'on trouvera un pivot dans chaque colonne *)
  for j=0 to n - 1 do
    let p = ref j in (* on cherche un pivot non nul dans la colonne j, sous la ligne j *)
    while !p < n && not (a.(!p).(j)) do
      incr p;
    done;
    (*  on place cette ligne à la ligne j,
        puis on l'ajoute aux autres si nécessaire *)
    echange a !p j;
    echange b !p j;
    for i = j+1 to (n-1) do 
      (* il n'est utile d'ajouter la ligne j à la ligne i que lorsque a_i,j est non nul *)
      if a.(i).(j) then ajoute a b i j;
    done;
  done;
  
  (* la matrice est maintenant sous forme échelonnée *)
  (* on remonte les colonnes pour calculer la solution x *)
  let x = Array.make n false in
  for j = n - 1 downto 0 do 
    x.(j) <- b.(j); (* la solution est imposée par b *)
    if x.(j) then (* on propage aux lignes d'au dessus si besoin *) 
      for i=0 to j - 1 do
        if a.(i).(j) then
          b.(i) <- not b.(i);
      done;
  done;
  x 

(* résolution du puzzle *)
let resout (g:grille) : case list =
  let n = taille g in
  let a = matrice_lumiere n in
  let b = vg n g in
  let x = pivot (n * n) a b in
  cases_vc n x


let q6 (n:int) (m:int) (p:int) : int =
  let g = gnmp n m p in
  let c = resout g in
  somme_ensemble c


(* affichage et valeurs numériques *)
let print_q6 (num:string) (n:int) (m:int) (p:int) : unit =
  Printf.printf "6.%s %d\n" num (q6 n m p);
  flush stdout

let qq6 () =
  print_q6 "a" 6 4 7;
  print_q6 "b" 31 4 7;
  print_q6 "c" 43 4 7;
  print_q6 "d" 52 4 7;
  Printf.printf "\n";
  flush stdout



(* -------------- *)
(* Question 10    *)
(* -------------- *)

(* calcule la table Eg *)
let forcebrute_une (g:grille) : bool array =
  let n = taille g in
  let h = vide n in (* grille de travail *)
  let eg = Array.make (n * n + 1) false in

  (* écrit dans eg.(d) "true" si il existe un choix pour les lignes <= i qui donnent
     une distance de d quand on choisit une case <= j sur la ligne i *)
  let rec aux i j =
    if i < 0 then
      let d = distance h g in
      eg.(d) <- true;
    else
      if j >= 0 then
        (appuie h (i,j);
         aux (i-1) (n-1);
         appuie h (i,j);
         aux i (j-1))
  in
  aux (n-1) (n-1);
  eg

let q10 (n:int) (m:int) (p:int) : int =
  let g = gnmp n m p in
  let eg = forcebrute_une g in
  let s = ref 0 in
  Array.iteri (fun d nn -> if nn then s := !s + d) eg;
  (!s mod 10000)


(* affichage et valeurs numériques *)
let print_q10 (num:string) (n:int) (m:int) (p:int) : unit =
  Printf.printf "10.%s %d\n" num (q10 n m p);
  flush stdout

let qq10 () =
  print_q10 "a" 3 1 2;
  print_q10 "b" 5 3 7;
  print_q10 "c" 6 3 7;
  Printf.printf "\n";
  flush stdout



(* -------------- *)
(* Question 11    *)
(* -------------- *)

let dyn (g:grille) : bool array =
  let n = taille g in
  (* on va calculer une table h de taille n * n * n * n^2 telle que
     h[i][j][k][d] sera le booléen :
     "est-il possible d'obtenir une distance d en ne prenant
      en compte que les lignes <= i, en appuyant sur les cases (i, j) et (i - 1, k) ?" *)
  (* la ligne i = 0 de h est inutilisée *)
  let h = Array.make_matrix n n (Array.make_matrix 0 0 false) in
  for i=0 to (n-1) do
    for j=0 to (n-1) do
      h.(i).(j) <- Array.make_matrix n (n*n + 1) false;
    done;
  done;

  (* pour calculer h, on utilise la relation suivante, pour i > 1 :
     h[i][j][k][d] =
         \/     h[i-1][k][l][d - (dligne i j) - (dint_haut i j k) - (dint_bas i j k l)]
     {0 <= l < n} 
     où dligne, dint_haut, dint_bas sont définis ci dessous *)       

  (* on écrira donc un algorithme dynamique, qui va calculer
     successivement toutes les cases de h[1], h[2],…, h[n-1] *)

  (* h[1] sera initialisée avec la relation suivante :
     h[1][j][k][d] = (d = dligne 1 j + dligne 0 k + dint_haut i j k + dint_haut 0 k j) *)
  
  
  (* dligne i j est la distance entre une ligne vide et
     la ligne i de g *après avoir appuyé sur (i,j)* *)
  let dligne = Array.make_matrix n n 0 in
  for i = 0 to n - 1 do
    let d = ref 0 in (* distance sans rien appuyer *)
    for l = 0 to n - 1 do 
      if est_allume g (i, l) then
        incr d;
    done;
    for j = 0 to n - 1 do (* pour chaque j, on corrige au voisinage de j *)
      dligne.(i).(j) <- !d;
      List.iter
        (fun l -> (* si (i, l) est censée être allumée : -1, sinon +1 *)
          if est_allume g (i, l) then
            dligne.(i).(j) <- dligne.(i).(j) - 1
          else
            dligne.(i).(j) <- dligne.(i).(j) + 1)
        (List.filter (fun l -> l >= 0 && l < n) [j - 1; j; j + 1]);
    done;
  done;
  let dligne i j = dligne.(i).(j) in
  
  (* dint_haut est la distance introduite sur la ligne i, lorsqu'on appuie sur (i,j),
     et que (i - 1, k) est la case appuyée ligne i - 1 *)
  let dint_haut i j k =
    if xor (k = j - 1 || k = j || k = j + 1) (est_allume g (i, k)) then
      -1
    else
      1
  in
  
  (* dint_bas est la distance introduite sur la ligne i-1, lorsqu'on appuie sur (i,j),
     et que (i - 1, k) et (i - 2, l) sont appuyées lignes i - 1, i - 2 *)
  let dint_bas i j k l =
    if xor (xor (j = k - 1 || j = k || j = k + 1) (j = l)) (est_allume g (i - 1, j)) then
      -1
    else
      1
  in


  (* initialisation de h[1] *)
  for j=0 to (n-1) do
    for k=0 to (n-1) do
      let dd = (dligne 1 j) + (dligne 0 k) + (dint_haut 1 j k) + (dint_haut 0 k j) in
      h.(1).(j).(k).(dd) <- true;
    done;
  done;
  
  (* remplissage de h *)
  for i=2 to (n-1) do
    for j=0 to (n-1) do
      for k=0 to (n-1) do
        for d=0 to (n*n) do
          let dd = d - (dligne i j) - (dint_haut i j k) in
          let l = ref 0 in
          while (!l < n && not h.(i).(j).(k).(d)) do
            let ddd = dd - (dint_bas i j k !l) in
            if ddd >= 0 && ddd <= n*n && h.(i-1).(k).(!l).(ddd) then 
              h.(i).(j).(k).(d) <- true;
            incr l;
          done;
        done;
      done;
    done;
  done;
  
  (* on retrouve Eg à partir de la ligne n - 1 de h,
     en prenant la disjonction pour tous les j, k. *)
  let eg = Array.make (n * n + 1) false in
  for d=0 to n*n do
    let j = ref 0 in
    while (!j < n && not eg.(d)) do
      let k = ref 0 in
      while (!k < n && not eg.(d)) do
        if h.(n-1).(!j).(!k).(d) then
          eg.(d) <- true;
        incr k;
      done;
      incr j;
    done;
  done;
  eg

let q11 (n:int) (m:int) (p:int) : int =
  let g = gnmp n m p in
  let eg = dyn g in
  let s = ref 0 in
  Array.iteri (fun d nn -> if nn then s := !s + d) eg;
  (!s mod 10000)


(* affichage et valeurs numériques *)
let print_q11 (num:string) (n:int) (m:int) (p:int) : unit =
  Printf.printf "11.%s %d\n" num (q11 n m p);
  flush stdout

let qq11 () =
  print_q11 "a" 12 3 7;
  print_q11 "b" 16 5 7;
  print_q11 "c" 20 3 7;
  Printf.printf "\n";
  flush stdout



(* ------------------------------- *)
(* Calcul de toutes les réponses   *)
(* ------------------------------- *)

let partie1 () =
  Printf.printf "u0 %d\n\n" u0;
  qq1 ();
  qq2 ();
  qq3 ();
  qq4 ();
  qq5 ();
  qq6 ()

let partie2 () =
  qq10 ();
  qq11 ()

let questions () =
  partie1 ();
  partie2 ()

let _ = questions ()
